package meerkat.voting.output;

import com.google.protobuf.ByteString;
import meerkat.protobuf.Crypto.EncryptionRandomness;
import meerkat.protobuf.Crypto.RandomnessGenerationProof;
import meerkat.protobuf.Voting.BallotSecrets;
import meerkat.protobuf.Voting.PlaintextBallot;
import meerkat.protobuf.Voting.SignedEncryptedBallot;
import meerkat.voting.gui.ui.controllersFX.VistaNavigator;
import meerkat.voting.output.outputcommands.AuditOutputCommand;
import meerkat.voting.output.outputcommands.CancelOutputCommand;
import meerkat.voting.output.outputcommands.CastOutputCommand;
import meerkat.voting.output.outputcommands.CommitOutputCommand;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A toy OutputDevice class
 * outputs everything simply to the System console
 */
public class FXWindowOutputDevice extends AsyncRunnableOutputDevice {

    private static final Logger logger = LoggerFactory.getLogger(FXWindowOutputDevice.class);

    public FXWindowOutputDevice() {
        super();
        logger.info("A FXWindowOutputDevice is constructed");
    }

    /**
     * Committing to the ballot.
     * Simply prints to the output stream all the details in the CommitOutputCommand.
     * @param command details to commit to, and the callback to call when finished
     */
    public void doCommitToBallot(CommitOutputCommand command) {
        logger.debug("entered method doCommitToBallot");
        PlaintextBallot plaintextBallot = command.getPlaintext();
        long plaintextSerialNumber = plaintextBallot.getSerialNumber();
        String toPrint = "";

        toPrint+="Commitment of Ballot #" + plaintextSerialNumber;
        toPrint+="\n";
        toPrint+="(channel): ";
        toPrint+="\n";
        toPrint+=bytesToString(command.getChannelIdentifierByteString());
        toPrint+="\n";
        toPrint+="(plaintext): ";
        toPrint+="\n";
        toPrint+=plaintextBallot;
        SignedEncryptedBallot signedEncryptedBallot = command.getSignedEncryptedBallot();
        long encryptedSerialNumber = signedEncryptedBallot.getEncryptedBallot().getSerialNumber();
        toPrint+="Commitment of Ballot #" + encryptedSerialNumber + " (ciphertext):";
        toPrint+="\n";
        if (plaintextSerialNumber != encryptedSerialNumber) {
            logger.error("plaintext and encryption serial numbers do not match!! plaintext# = " +
                    plaintextSerialNumber + ", ciphertext# = " + encryptedSerialNumber);
        }
        ByteString encryptedData = signedEncryptedBallot.getEncryptedBallot().getData().getData();
        toPrint+=bytesToString(encryptedData);
        toPrint+="\n";
        VistaNavigator.printFX(toPrint, 'w');
        command.getCallback().onSuccess(null);
    }


    /**
     * auditing the ballot.
     * prints to the output stream the ballot secrets (the encryption randomness and its proof of random generation)
     * @param command An auditing command with the callback to finally call
     */
    public void doAudit(AuditOutputCommand command) {
        logger.debug("entered method doAudit");
        try {
            Thread.sleep(500);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        String msg = "Auditing";
        VistaNavigator.printFX(msg, 'a');
        BallotSecrets ballotSecrets = command.getBallotSecrets();
        printEncryptionRandomness(ballotSecrets.getEncryptionRandomness());
        printRandomnessGenerationProof (ballotSecrets.getProof());
        command.getCallback().onSuccess(null);
    }

    /**
     * Casting the ballot (actually does nothing new)
     * @param command a CastOutputCommand with the details and the callback
     */
    public void doCastBallot(CastOutputCommand command) {
        logger.debug("entered method doCastBallot");
        String msg = "Ballot finalized for casting!";
        VistaNavigator.printFX(msg, 'a');
        command.getCallback().onSuccess(null);
    }


    /**
     * Canceling the ballot (actually does nothing new)
     * @param command a CancelOutputCommand with the details and the callback
     */
    public void doCancel(CancelOutputCommand command) {
        logger.debug("entered method doCancel");
        System.out.println("Ballot cancelled!");
        command.getCallback().onSuccess(null);
    }


    private void printEncryptionRandomness (EncryptionRandomness encryptionRandomness) {
        String msg = "Encryption Randomness = ";
        ByteString data = encryptionRandomness.getData();
        msg+=bytesToString(data);
        VistaNavigator.printFX(msg, 'a');
    }

    private void printRandomnessGenerationProof (RandomnessGenerationProof proof) {
        String msg = "Proof of randomness generation:";
        ByteString data = proof.getData();
        msg+=bytesToString(data);
        VistaNavigator.printFX(msg, 'a');
    }


    /*
     * Returns the UTF8 decoding of byte-string data
     */
    private static String bytesToString(ByteString data) {
        return data.toStringUtf8();
    }

}
